/*
   input.c -- PIKT alarm/script input routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/input.c,v 1.15 2004/01/29 15:06:11 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"

#include "piktd.h"

#ifndef SEEK_SET
#define	SEEK_SET 0
#endif
#ifndef SEEK_CUR
#define	SEEK_CUR 1
#endif
#ifndef SEEK_END
#define	SEEK_END 2
#endif

/* in the case of NOINPUT, do nothing; in the case of PROCINPUT, open the
   process for reading, piping through a filter, if specified; in the case
   of LOGINPUT, open the log file for reading, based on the history value,
   if available, determine the point just beyond the last recorded file
   position, exec() a tail of the log file, piping through a filter, if
   specified; in the case of FILEINPUT, open the file for reading, piping
   through a filter, if specified
*/
int
openinput (FILE **infile, pid_t *inpid, int inputtype, char *inputsource, FILE **tmpfile, char *inputtmpfil)
{
	static char fnnam[] = "openinput";
	char inputcmd[MAXCMD+1];
	char *sc;
	struct stat fbuf;
	unsigned long poscurr;
	int tmpfilerr;

	tmpfilerr = FALSE;
	if (inputsource && (strlen(inputsource) > MAXCMD)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_INPUT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, input exceeds %d char limit", ALMNAM, MAXCMD);
		STATUS = INACTIVE;
		return ERROR;
	}
	if ((inputtype != NOINPUT) && ! strlen(inputsource)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_NO_INPUT_SOURCE, __FILE__, __LINE__, fnnam, FALSE, "in %s, no input source", ALMNAM);
			STATUS = INACTIVE;
			return ERROR;
	}
	inputino = 0;
	inputsize = 0;
	inputposend = 0;
	switch (inputtype) {
		case NOINPUT:
			*infile = (FILE *) NULL;
			*inpid = 0;
			break;
		case PROCINPUT:
			sprintf(inputcmd, "%s%s%s", inputsource, FILTER ? " | " : nullstring, FILTER ? FILTER : nullstring);
			sc = stripesc(inputcmd);
			if ((*infile = inputopen(sc, inpid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_INPUT_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, inputopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			break;
		case LOGINPUT:
			if (stat(inputsource, &fbuf) != 0) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			inputino = fbuf.st_ino;
			inputsize = fbuf.st_size;
			if ((*infile = fopen(inputsource, "r")) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fopen() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			/* if this is the same log file as before,
			   and it has not shrunk (been zeroed out)
			 */
			if ((inputino == (ino_t)INPUTLOGINO) && (inputsize >= (off_t)INPUTLOGSIZE)) {
			/* but what about case where we have zeroed out
			   log file and it has since grown to be larger
			   than last alert run?  an unusual case, but it
			   could happen!
			 */
				/* if the previous file was not zero length,
				   attempt to position to previous eof pos
				 */
				if (INPUTLOGPOSEND) {
					if (fseek(*infile, (long)INPUTLOGPOSEND, SEEK_SET)) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_SEEK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fseek() error for %s", ALMNAM, inputsource);
						rewind(*infile);
						if (fclose(*infile))
							logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for %s", ALMNAM, inputsource);
						STATUS = INACTIVE;
						return ERROR;
					}
				}
			}
			/* determine current position in file for passing
			   to tail command
			 */
			if ((poscurr = ftell(*infile)) == -1L) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_TELL_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, ftell() error for %s", ALMNAM, inputsource);
				if (fclose(*infile))
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			/* determine eof position for this file */
			if (fseek(*infile,0L,SEEK_END)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_SEEK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fseek() error for %s", ALMNAM, inputsource);
				if (fclose(*infile))
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			} else {
				if ((inputposend = ftell(*infile)) == -1L) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_TELL_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, ftell() error for %s", ALMNAM, inputsource);
					if (fclose(*infile))
						logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for %s", ALMNAM, inputsource);
					STATUS = INACTIVE;
					return ERROR;
				}
			}
			if (fclose(*infile)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			sprintf(inputcmd, "%s +%ld%c %s%s%s", tailcmd, poscurr+1, 'c', inputsource, FILTER ? " | " : nullstring, FILTER ? FILTER : nullstring);
			sc = stripesc(inputcmd);
			if ((*infile = inputopen(sc, inpid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_INPUT_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, inputopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			break;
		case FILEINPUT:
			if (stat(inputsource, &fbuf) != 0) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			sprintf(inputcmd, "/bin/cat %s%s%s", inputsource, FILTER ? " | " : nullstring, FILTER ? FILTER : nullstring);
			sc = stripesc(inputcmd);
			if ((*infile = inputopen(sc, inpid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_INPUT_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, inputopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			if (inputtmpfil) {
				if ((*tmpfile = fopen(inputtmpfil, "w")) == NULL) {
					tmpfilerr = TRUE;
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fopen() error for %s", ALMNAM, inputtmpfil);
				}
				if (chmod(inputtmpfil, fbuf.st_mode)) {
					tmpfilerr = TRUE;
					logreturn(PIKTLOG, LOG_ERR, ERRID_CHMOD_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, chmod() error for %s", ALMNAM, inputtmpfil);
				}
				if (chown(inputtmpfil, fbuf.st_uid, fbuf.st_gid)) {
					tmpfilerr = TRUE;
					logreturn(PIKTLOG, LOG_ERR, ERRID_CHOWN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, chown() error for %s", ALMNAM, inputtmpfil);
				}
				if (tmpfilerr) {
					inputclose(*infile, *inpid);
					STATUS = INACTIVE;
					return ERROR;
				}
			} else {
				*tmpfile = (FILE *) NULL;
			} 
			break;
	}
	return 0;
}

/* close the input file handle, if called for
*/
int
closeinput (FILE **infile, pid_t inpid, FILE **tmpfile)
{
	static char fnnam[] = "closeinput";

	if (*infile)
		inputclose(*infile, inpid);
	*infile = (FILE *) NULL;
	if (*tmpfile) {
		if (fclose(*tmpfile))
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error", ALMNAM);
	}
	*tmpfile = (FILE *) NULL;
	return 0;
}

/* spawn a child process; the child opens the specified input, whether a
   process accessing some file or some other process; the process output,
   both stdout and stderr, is fed back to the parent; doing it this way,
   and by not using a simple popen(), we can capture output to both
   stdout and stderr, giving us useful debugging info that might otherwise
   be lost; on some machines, we coordinate the parent and child via IPC;
   on other machines, we just have the parent sleep() an arbitary amount
   while waiting for the child
 */
FILE *
inputopen (char *cmdstr, pid_t *pid)
{
	static char fnnam[] = "inputopen";
	int fd[2];
	FILE *fp;
	FILE *logfile;
	int logfd;

	fp = NULL;
	if (pipe(fd) < 0)
		logexit(PIKTLOG, LOG_ERR, ERRID_PIPE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, inputopen(), pipe() error", ALMNAM);
	ipcopen(0, __FILE__, __LINE__);
	switch (*pid = fork()) {
		case -1:	/* error */
			logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fork() error", ALMNAM);
			break;
		/* pipe from child to parent */
		case 0:		/* child */
			close(fd[0]);
			logfile = fopen(piktlog, "a");
			logfd = fileno(logfile);
			dup2(logfd, STDERR_FILENO);
			if (fd[1] != STDOUT_FILENO) {
				dup2(fd[1], STDOUT_FILENO);
				close(fd[1]);
			}
			wait_parent(0, __FILE__, __LINE__);
			tell_parent(0, __FILE__, __LINE__);
			execl(SHELL, "sh", "-c", cmdstr, (char *) NULL);
			fclose(logfile);
			exit(0);
			break;
		default:	/* parent */
			close(fd[1]);
			tell_child(0, __FILE__, __LINE__);
			wait_child(0, __FILE__, __LINE__);
			sleep(fork_delay);
			if ((fp = fdopen(fd[0], "r")) == NULL)
				logreturn(PIKTLOG, LOG_ERR, ERRID_FDOPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fdopen() error", ALMNAM);
	}
	return fp;
}

/* close the input stream
 */
void
inputclose (FILE *fp, pid_t pid)
{
	static char fnnam[] = "inputclose";

	if (fclose(fp))
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fclose() error", ALMNAM);
	ipcclose(0, __FILE__, __LINE__);
	return;
}
